/**
 * This file is part of the OpenVIDIA project at http://openvidia.sf.net
 * Copyright (C) 2004, James Fung
 *
 * OpenVIDIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * OpenVIDIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenVIDIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **/

#include "GLDisplay.h"

#include <Cg/cgGL.h>
//#include <glh/glh_extensions.h>

#include <iostream>
using namespace std;
#define TEXTURE_WIDTH   128 
#define TEXTURE_HEIGHT   128 

GLDisplay::GLDisplay(int win)  : GenericDisplay(win)  {
  // create vector to hold textures
  texNames = new GLuint[1];
  cout << "Constructed GLDisplay" << endl;
}

void GLDisplay::showstats() {}

void GLDisplay::initGL() {
  glClearColor (0.0, 0.0, 0.0, 0.0);
  glShadeModel(GL_FLAT);
  glEnable(GL_DEPTH_TEST);
  // rosco
  glGenTextures(1,texNames);
}


// should bounds check on id?
void GLDisplay::init_texture(int id, int width, int height, void *data) {
  glBindTexture(GL_TEXTURE_2D, texNames[0]);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 
	       width, height, 
	       0, GL_RGB, GL_UNSIGNED_BYTE, data);
}

//bind textures to texturing units
void GLDisplay::bindTextures() {
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   glEnable(GL_TEXTURE_2D);

   glBindTexture(GL_TEXTURE_2D, texNames[0]);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
}

void GLDisplay::idleFunc() {
  bindTextures();
  render();
}

void GLDisplay::render() {
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glEnable(GL_TEXTURE_2D);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
  glBindTexture(GL_TEXTURE_2D, texNames[0]);
  
  glBegin(GL_QUADS);
  int dist=0;
  glTexCoord2f(0.0, 1.0); glVertex3f(-2.0, -2.0, dist);
  glTexCoord2f(0.0, 0.0); glVertex3f(-2.0, 0.0, dist);
  glTexCoord2f(1.0, 0.0); glVertex3f(0.0, 0.0, dist);
  glTexCoord2f(1.0, 1.0); glVertex3f(0.0, -2.0, dist);
  
  glTexCoord2f(0.0, 1.0); glVertex3f(-2.0,0.0, dist);
  glTexCoord2f(0.0, 0.0); glVertex3f(-2.0, 1.0, dist);
  glTexCoord2f(1.0, 0.0); glVertex3f(-1.0, 1.0, dist);
  glTexCoord2f(1.0, 1.0); glVertex3f(-1.0, 0.0, dist);
  
  glTexCoord2f(0.0, 1.0); glVertex3f(-2.0,1.0, dist);
  glTexCoord2f(0.0, 0.0); glVertex3f(-2.0, 1.5, dist);
  glTexCoord2f(1.0, 0.0); glVertex3f(-1.5, 1.5, dist);
  glTexCoord2f(1.0, 1.0); glVertex3f(-1.5, 1.0, dist);
  glEnd();

  //common
  glFlush();
  glDisable(GL_TEXTURE_2D);
}

void GLDisplay::grabDisplay( float *Dm, float *Dn, float *Dt) {
  cerr << "GL version of this grabDisplay not implemented FOOL" << endl;
}

void GLDisplay::grabDisplay( unsigned char *Dm, unsigned char *Dn, unsigned char *Dt) {
  cerr << "GL version of this grabDisplay not implemented FOOL" << endl;
}
