/**
 * This file is part of the OpenVIDIA project at http://openvidia.sf.net
 * Copyright (C) 2004, James Fung
 *
 * OpenVIDIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * OpenVIDIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenVIDIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **/

#ifndef _UDAT_INCLUDE
#define _UDAT_INCLUDE
#include <iostream>
#include "GenericCapture.h"
using namespace std;


///Udat is a class encapsulating interface with a U.dat U matrix file 
///for eigen image recognition.  The U.dat should be generated by the
///find_matrix of find_matrix_gs program
class Udat  : public GenericCapture
{
  ///file pointer for to the U.dat file. Used only on opening.
  FILE *fp;
  ///filename of the U matrix data file.  Defaults to U.dat
  char *fname;
  int UdatWidth;
  int UdatHeight;
  ///Memory where the data is stored.
  double *store;
  bool isStored;

  ///remember place for GenericCapture interface
  int currentcol;

  ///if singular values are needed, store them here
  double *sing_values;
  int num_sing_values;
  public:
    Udat(char *Udat_filename);
    ~Udat(){};
    int getWidth();
    int getHeight();
    double getColNorm(int column);
    int getRGBACols(int startcol, float *rgba_buffer);
    int getRGBCols(int startcol, float *rgba_buffer);
    bool freeUdat();
    void readSVdat(char *);

    //Override GenericCapture interface
    void *getRGBData();
    void *getRGBAData();
    void *getRGBAData4f();
    int getRGBWidth();
    int getRGBHeight();
    void advanceFrame();
    void advanceFrame(int);
    bool trySingVals();
    float *getColumnData_1f(int col);
    float *getColumnData_1f();
    float getSingVal(int);
};

#endif
