/**
 * This file is part of the OpenVIDIA project at http://openvidia.sf.net
 * Copyright (C) 2004, James Fung
 *
 * OpenVIDIA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * OpenVIDIA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenVIDIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * based on Video4linux2 capture.c example code.
 **/

#ifndef _ONV_V4L2_H
#define _ONV_V4L2_H
#include "GenericCapture.h"
#include "GenericDisplay.h"
#include <cc++2/cc++/thread.h>
//#include <thread.h>
#include <cc++2/cc++/config.h>
#include <cc++2/cc++/exception.h>
#include <pthread.h>
#include <semaphore.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>
#include <asm/types.h>          /* for videodev2.h */
#include <linux/videodev2.h>

typedef enum {
	IO_METHOD_READ,
	IO_METHOD_MMAP,
	IO_METHOD_USERPTR,
} io_method;

struct buffer {
        void *                  start;
        size_t                  length;
};

using namespace std;
using namespace ost;

//class V4L2 : public Thread, public Mutex, public VidCap
class V4L2 : public Thread, public Mutex
{
 private:
  int bufferUsed;
  char *dev_name;
  io_method io;
  int fd;
  struct buffer *buffers;
  unsigned int n_buffers;
  char *read_frame(void);
  void mainloop(void);
  void start_capturing(void);
  void stop_capturing(void);
  void uninit_device(void);
  void init_mmap(void);
  void init_read(unsigned int);
  void init_userp(unsigned int);
  void init_device(void);
  void close_device(void);
  void open_device(void);
  void set_controls();

 protected:
  const static int DefaultCaptureWidth=320;
  const static int DefaultCaptureHeight=240;
  int CaptureWidth,CaptureHeight;

 public:
  V4L2();
  ~V4L2();
  // Start Thread
  void run();
  void tellThreadDoneWithBuffer();
  void lock() { ENTER_CRITICAL; }
  void unlock() { LEAVE_CRITICAL; }
};

#endif

